% Version 1.000
%
% Code provided by Hugo Larochelle and Ruslan Salakhutdinov
%
% Permission is granted for anyone to copy, use, modify, or distribute this
% program and accompanying programs and documents for any purpose, provided
% this copyright notice is retained and prominently displayed, along with
% a note saying that the original programs are available from our
% web page.
% The programs and documents are distributed without any warranty, express or
% implied.  As the programs were written for research purposes only, they have
% not been tested to the degree that would be advisable in any important
% application.  All use of these programs is entirely at the user's own risk.

close all 
if restart ==1,

  if exist('epsilon_mf','var')
    epsilonw      = epsilon_mf;
    epsilonvb     = epsilon_mf; 
    epsilonhb     = epsilon_mf;
  else
    epsilonw      = 0.001;   % Learning rate for weights 
    epsilonvb     = 0.001;   % Learning rate for biases of visible units 
    epsilonhb     = 0.001;   % Learning rate for biases of hidden units 
  end

  weightcost  = 0.0002;   
  initialmomentum  = 0.5;
  finalmomentum    = 0.9;

  [numcases numdims numbatches]=size(batchdata);

  numdim=numdims;

  restart=0;
  epoch=1;
  % Initializing symmetric weights and biases.
  
  vishid     = 0.001*randn(numdim, numhid);
  hidpen     = 0.01*randn(numhid,numpen); 

  hidbiases  = zeros(1,numhid);
  visbiases  = zeros(1,numdim);
  penbiases  = zeros(1,numpen);

  poshidprobs = zeros(numcases,numhid);
  neghidprobs = zeros(numcases,numhid);
  posprods    = zeros(numdim,numhid);
  negprods    = zeros(numdim,numhid);

  
  vishidinc  = zeros(numdim,numhid);
  hidpeninc  = zeros(numhid,numpen);  

  hidbiasinc = zeros(1,numhid);
  visbiasinc = zeros(1,numdim);
  penbiasinc = zeros(1,numpen);

  %%%% This code also adds sparcity penalty 
  sparsetarget = .2;
  sparsetarget2 = .1;
  sparsecost = .001;
  sparsedamping = .9;

  hidbiases  = 0*log(sparsetarget/(1-sparsetarget))*ones(1,numhid);
  hidmeans = sparsetarget*ones(1,numhid);
  penbiases  = 0*log(sparsetarget2/(1-sparsetarget2))*ones(1,numpen);
  penmeans = sparsetarget2*ones(1,numpen);

  load fullpo.mat

  hidpen = vishid;
  penbiases = hidbiases;
  visbiases_l2 = visbiases;

  load fullvh.mat
  hidbiases = (hidbiases + visbiases_l2);
  epoch=1; 

  neghidprobs = (rand(numcases,numhid));
  data = round(rand(numcases,numdims));
  neghidprobs = 1./(1 + exp(-data*(2*vishid) - repmat(hidbiases,numcases,1)));

  epsilonw      = epsilonw/(1.000015^((epoch-1)*600));
  epsilonvb      = epsilonvb/(1.000015^((epoch-1)*600));
  epsilonhb      = epsilonhb/(1.000015^((epoch-1)*600));

  tot = 0; 

  % Mean-field predictor initialization
  mf_predictor_init;
end

dbm_mf_tot_toc = 0;

for epoch = epoch:maxepoch 
  tic;
  [numcases numdims numbatches]=size(batchdata);

  if epoch>5,
    momentum=finalmomentum;
  else
    momentum=initialmomentum;
  end;

  fprintf(1,'epoch %d \t eps %f\r',epoch,epsilonw); 
  errsum=0;

  [numcases numdims numbatches]=size(batchdata);

  rr = randperm(numbatches);  
  batch=0; 
  mf_it=0;
  mf_err_h1=0;
  mf_err_h2=0;
  for batch_rr = rr; 
    batch=batch+1;  
    fprintf(1,'epoch %d batch %d\r',epoch,batch); 
    tot=tot+1; 
    epsilonw = max(epsilonw/1.000015,0.00010);
    epsilonvb = max(epsilonvb/1.000015,0.00010);
    epsilonhb = max(epsilonhb/1.000015,0.00010);


    %%%%%%%%% START POSITIVE PHASE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    data = batchdata(:,:,batch);
    targets = batchtargets(:,:,batch); 
    data = double(data > rand(numcases,numdim));  

    %%%%% First fo MF %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % Get prediction for mf parameters
    mf_prediction;

    [poshidprobs, pospenprobs, tot_it] = ...
	mf_class_init(data,vishid,hidbiases,visbiases,hidpen,penbiases,init_h1,init_h2,maxit);

    % Update mf predictor
    mf_prediction_update;

    ff = find(init_h1==1); 
    init_h1(ff) = 1-eps; 
    ff = find(init_h1==0); 
    init_h1(ff) = eps; 
    ff = find(init_h2==1); 
    init_h2(ff) = 1-eps; 
    ff = find(init_h2==0); 
    init_h2(ff) = eps; 

    err_h1 = -sum(sum(poshidprobs.*log(init_h1)+(1-poshidprobs).*log((1-init_h1)),2))/(numcases*numhid);
    err_h2 = -sum(sum(pospenprobs.*log(init_h2)+(1-pospenprobs).*log((1-init_h2)),2))/(numcases*numpen);

    err_h1 = err_h1 + sum(sum(poshidprobs.*log(poshidprobs+eps*(poshidprobs==0))+(1-poshidprobs).*log((1-poshidprobs)+eps*(poshidprobs==1)),2))/(numcases*numhid);
    err_h2 = err_h2 + sum(sum(pospenprobs.*log(pospenprobs+eps*(pospenprobs==0))+(1-pospenprobs).*log((1-pospenprobs)+eps*(pospenprobs==1)),2))/(numcases*numpen);

    mf_it = mf_it + tot_it;
    mf_err_h1 = mf_err_h1 + err_h1;
    mf_err_h2 = mf_err_h2 + err_h2;

    bias_hid= repmat(hidbiases,numcases,1);
    bias_pen = repmat(penbiases,numcases,1);
    bias_vis = repmat(visbiases,numcases,1);
    
    posprods    = data' * poshidprobs;
    posprodspen = poshidprobs'*pospenprobs;

    poshidact   = sum(poshidprobs);
    pospenact   = sum(pospenprobs);
    posvisact = sum(data);


    %%%%%%%%% END OF POSITIVE PHASE  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    negdata_CD1 = 1./(1 + exp(-poshidprobs*vishid' - bias_vis));


    %%%%% START NEGATIVE PHASE  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    for iter=1:5 
      neghidstates = neghidprobs > rand(numcases,numhid);
      
      negpenprobs = 1./(1 + exp(-neghidstates*hidpen - bias_pen));
      negpenstates = negpenprobs > rand(numcases,numpen);

      negdataprobs = 1./(1 + exp(-neghidstates*vishid' - bias_vis));
      negdata = negdataprobs > rand(numcases,numdim);

      totin = negdata*vishid + bias_hid + negpenstates*hidpen';
      neghidprobs = 1./(1 + exp(-totin));

    end 
    negpenprobs = 1./(1 + exp(-neghidprobs*hidpen - bias_pen));

    negprods  = negdata'*neghidprobs;
    negprodspen = neghidprobs'*negpenprobs;
    neghidact = sum(neghidprobs);
    negpenact   = sum(negpenprobs);
    negvisact = sum(negdata); 


    %%%%%%%%% END OF NEGATIVE PHASE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    err= sum(sum( (data-negdata_CD1).^2 ));
    errsum = err + errsum;

    %%%%%%%%% UPDATE WEIGHTS AND BIASES %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

    visbiasinc = momentum*visbiasinc + (epsilonvb/numcases)*(posvisact-negvisact);

    hidmeans = sparsedamping*hidmeans + (1-sparsedamping)*poshidact/numcases;
    sparsegrads = sparsecost*(repmat(hidmeans,numcases,1)-sparsetarget);

    penmeans = sparsedamping*penmeans + (1-sparsedamping)*pospenact/numcases;
    sparsegrads2 = sparsecost*(repmat(penmeans,numcases,1)-sparsetarget2);


    vishidinc = momentum*vishidinc + ...
	epsilonw*( (posprods-negprods)/numcases - weightcost*vishid - ...   
		   data'*sparsegrads/numcases );
    hidbiasinc = momentum*hidbiasinc + epsilonhb/numcases*(poshidact-neghidact) ...
	-epsilonhb/numcases*sum(sparsegrads);

    hidpeninc = momentum*hidpeninc + ...
	epsilonw*( (posprodspen-negprodspen)/numcases - weightcost*hidpen - ...
		   poshidprobs'*sparsegrads2/numcases - (pospenprobs'*sparsegrads)'/numcases );
    penbiasinc = momentum*penbiasinc + epsilonhb/numcases*(pospenact-negpenact) ...
	-epsilonhb/numcases*sum(sparsegrads2);


    vishid = vishid + vishidinc;
    hidpen = hidpen + hidpeninc;
    visbiases = visbiases + visbiasinc;
    hidbiases = hidbiases + hidbiasinc;
    penbiases = penbiases + penbiasinc;
    %%%%%%%%%%%%%%%% END OF UPDATES %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
    %if rem(batch,50)==0  
    %  figure(1);
    %  dispims(negdata',28,28);
    %
    %  % figure(2);
    %  %  subplot(2,2,1);
    %  %    hist(vishid(:),100);
    %  %  subplot(2,2,2);
    %  %    hist(vishidinc(:),100);
    %  %  subplot(2,2,3);
    %  %    hist(hidpen(:),100);
    %  %  subplot(2,2,4);
    %  %    hist(hidpeninc(:),100);
    %
    %
    %
    %  drawnow
    %end  

  end
  fprintf(1, ['Average number of MF iterations %6.1f\n  MF prediction errors' ...
	      ' h1=%f, h2=%f \n\n'], mf_it/numbatches, mf_err_h1/numbatches, mf_err_h2/numbatches); 

  save  full_dbm hidpen penbiases vishid hidbiases visbiases epoch;
  save_mf_predictor;
  dbm_mf_toc = toc;
  dbm_mf_tot_toc = dbm_mf_tot_toc + dbm_mf_toc;
end;




